/*++

Copyright (c) 2004 Alexander A. Telyatnikov (Alter)

Module Name:
    DbgPrintLog.cpp

Abstract:

Author:
    Alexander A. Telyatnikov (Alter)

Environment:
    User mode only

Notes:

    THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
    IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
    OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
    IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
    INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
    NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
    THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Revision History:

--*/

#include <windows.h>
#include <stdio.h>
#include "..\Driver\DbgPrnHk.h"
#include "SvcManLib.h"
#include "Privilege.h"
#include "flusher.h"

typedef struct _DBG_OUTPUT_DEBUG_STRING_BUFFER {
    ULONG ProcessId;
    UCHAR Msg[4096-sizeof(ULONG)];
} DBG_OUTPUT_DEBUG_STRING_BUFFER, *PDBG_OUTPUT_DEBUG_STRING_BUFFER;

PWCHAR LogFile = L"DbgPrint.log";  // default log-file name
WCHAR LogFileN[MAX_PATH*2];
WCHAR LogFileSuf[MAX_PATH];

BOOLEAN esc_key     = FALSE;
BOOLEAN pause_key   = FALSE;
BOOLEAN next_log    = FALSE;
BOOLEAN flush_file  = FALSE;

BOOLEAN log_pid  = FALSE;
BOOLEAN log_tid  = FALSE;
BOOLEAN log_mode = FALSE;
BOOLEAN log_irql = FALSE;

BOOLEAN skip_kernel = FALSE;
BOOLEAN skip_umode  = FALSE;

BOOLEAN stdout_mode = FALSE;
BOOLEAN sync_mode   = FALSE;
BOOLEAN stdout_copy = FALSE;

PDbgPrnHk_GetMessages_USER_OUT MsgBuffer = NULL;

// dbg messages capture buffers for user-mode
// (via OutputDebugString)
// '2' suffix  - buffer/counter for reading from get_messages()
// '2a' suffix - buffer/counter for writing from get_ODS_messages()
HANDLE hODS_lock = NULL;
ULONG MsgCount2 = 0;
ULONG MsgCount2a = 0;
ULONG MaxLogs = -1;
PDbgPrnHk_GetMessages_USER_OUT MsgBuffer2 = NULL;
PDbgPrnHk_GetMessages_USER_OUT MsgBuffer2a = NULL;

// settings
ULONG BufferSize = 1024*1024;

ULONG MaxLogSize = 1024*1024*128; //128Mb

ULONG Start_i = 0;

ULONG MaxSleepCount = 5;

HANDLE StdIn  = NULL;
HANDLE StdOut = NULL;

VOID
TeeOutput(
    HANDLE h,
    PCHAR Buffer,
    ULONG Length
    )
{
    WriteFileBuffer(h, Buffer, Length);
    if(!stdout_copy || stdout_mode) {
        return;
    }
    WriteConsole(StdOut, Buffer, Length, &Length, NULL);
} // end TeeOutput()

/*
 Catch keyboard messages
*/
DWORD
__stdcall
get_kbd(PVOID ctx)
{
    ULONG key = 0;
    ULONG tmp;
    INPUT_RECORD inp;

    while(!esc_key) {
        if(GetNumberOfConsoleInputEvents(StdIn, &key) && key) {
            key = ReadConsoleInput(StdIn, &inp, 1, &tmp);
            if(key &&
               inp.EventType == KEY_EVENT &&
               inp.Event.KeyEvent.bKeyDown) {
                switch(inp.Event.KeyEvent.wVirtualScanCode) {
                case 0x01: /*VK_ESC*/
                    esc_key = TRUE;
                    break;
                case 0x39: /*VK_SPACE*/
                    pause_key = !pause_key;
                    printf("Log %s\n", !pause_key ? "restarted" : "paused");
                    break;
                case 0x2e: /*VK_C*/
                    if(stdout_mode)
                        break;
                    stdout_copy = !stdout_copy;
                    break;
                case 0x31: /*VK_N*/
                    if(!stdout_mode)
                        next_log = TRUE;
                    break;
                case 0x1f: /*VK_S*/
                    sync_mode = !sync_mode;
                    break;
                case 0x21: /*VK_F*/
                    flush_file = TRUE;
                    break;
                case 0x25: /*VK_K*/
                    skip_kernel = !skip_kernel;
                    printf("kernel messages %s\n", !skip_kernel ? "on" : "off");
                    break;
                case 0x16: /*VK_Y*/
                    skip_umode = !skip_umode;
                    printf("user-mode messages %s\n", !skip_umode ? "on" : "off");
                    break;
                }
            }
        }
        Sleep(100);
    }
    return 0;
} // end get_kbd()


/*
 Catch strings from OutputDebugString()
*/
DWORD
__stdcall
get_ODS_messages(PVOID ctx)
{
    HANDLE hCompleteEvent;
    HANDLE hNewMsgEvent;
    HANDLE hdbgf;
    PDBG_OUTPUT_DEBUG_STRING_BUFFER ODSmem;
    ULONG j = 0;
    ULONG status;

    SECURITY_ATTRIBUTES SecAttr;
    SECURITY_DESCRIPTOR SecDesc;

    // Connect to user-mode logger
    memset(&SecAttr, 0, sizeof(SecAttr));
    memset(&SecDesc, 0, sizeof(SecDesc));
    SecAttr.nLength = sizeof(SECURITY_ATTRIBUTES);
    SecAttr.bInheritHandle = TRUE;
    SecAttr.lpSecurityDescriptor = &SecDesc;

    if(!InitializeSecurityDescriptor(&SecDesc, SECURITY_DESCRIPTOR_REVISION)) {
        printf("can't init SecurityDescriptor, (%#x)\n",
            GetLastError());
        return -1;
    }

    if(!SetSecurityDescriptorDacl(&SecDesc, TRUE, NULL, FALSE)) {
        printf("can't set SD Dacl, (%#x)\n",
            GetLastError());
        return -1;
    }

    hNewMsgEvent = CreateEvent(&SecAttr, FALSE, FALSE, "DBWIN_DATA_READY");

    if(!hNewMsgEvent) {
        printf("can't create sync event, (%#x)\n",
                GetLastError());
        return -1;
    }

    hCompleteEvent = CreateEvent(&SecAttr, FALSE, FALSE, "DBWIN_BUFFER_READY");

    if(!hCompleteEvent) {
        printf("can't create sync event (2), (%#x)\n",
                GetLastError());
        return -1;
    }

    hdbgf = CreateFileMapping((HANDLE)(-1),
                        &SecAttr, PAGE_READWRITE,
                        0, 4096,
                        "DBWIN_BUFFER");

    if (!hdbgf) {
        printf("can't create file mapping, (%#x)\n",
                GetLastError());
        return -1;
    }

    ODSmem = (PDBG_OUTPUT_DEBUG_STRING_BUFFER)MapViewOfFile(hdbgf, FILE_MAP_READ, 0, 0, 512);

    if (!ODSmem) {
        printf("can't map shared mem (%#x)\n",
                GetLastError());
        return -1;
    }

    // send Ready message
    SetEvent(hCompleteEvent);

    while(!esc_key) {
        // wait for incoming messages
        status = WaitForSingleObject(hNewMsgEvent, INFINITE);
        if(status == WAIT_OBJECT_0) {
            WaitForSingleObject(hODS_lock, INFINITE);

            // copy message to buffer in unified format
            j = MsgCount2a;
            QueryPerformanceCounter((PLARGE_INTEGER)&MsgBuffer2a[j].TimeStamp);
            memcpy(&(MsgBuffer2a[j].Msg[0]), &(ODSmem->Msg[0]), sizeof(ODSmem->Msg));
            MsgBuffer2a[j].Length = strlen((CCHAR*)&(ODSmem->Msg[0]));
            MsgBuffer2a[j].ProcessId = (PVOID)(ODSmem->ProcessId);
            MsgBuffer2a[j].ThreadId = (PVOID)(-1);
            MsgBuffer2a[j].Irql = 0;
            j++;
            MsgCount2a = j;

            ReleaseMutex(hODS_lock);

            SetEvent(hCompleteEvent);
        }
    }
    return 0;
} // end get_ODS_messages()

static const char help_text2[] = 
    "Commands:\n"
    "\t'Esc'   - exit\n"
    "\t'N'     - start new log\n"
    "\t'F'     - flush log buffer\n"
    "Mode switchers (toggle on'off):\n"
    "\t'Space' - pause\n"
    "\t'S'     - synchronous mode\n"
    "\t'C'     - copy log to stdout\n"
    "\t'K'     - capture kernel messages\n"
    "\t'U'     - capture user-mode messages\n"
    ;
/*
 merge user- and kernel-mode logs
*/
void
get_messages()
{
    HANDLE h;
    HANDLE fh;

    ULONG len;
    ULONG i, j, j2;
    WCHAR a;
    PWCHAR NumPos;
    ULONG returned;
    BOOLEAN new_line;
    CHAR DbgPrintBuf[256];
    ULONG SleepCount;
    CHAR tab3[] = "\t\t\t";
    ULONG CurLogSize;
    SYSTEMTIME SysTime;
    OSVERSIONINFO OsVer;
    PCHAR PlatformStr;

    PDbgPrnHk_GetMessages_USER_OUT tmp_MsgBuffer2;
    DbgPrnHk_GetVersion_USER_OUT VerBuffer;

    BOOLEAN last2 = FALSE;
    PDbgPrnHk_GetMessages_USER_OUT CurMsg = NULL;

    // Connect to driver
    h = CreateFile(NT_DbgPrnHk_USER_NAME, GENERIC_READ | GENERIC_WRITE,
                   /*FILE_SHARE_READ | FILE_SHARE_WRITE*/ 0,
                   NULL,
                   OPEN_EXISTING,
                   FILE_ATTRIBUTE_NORMAL/* | FILE_FLAG_NO_BUFFERING*/,  NULL);
    if(!h || h == (HANDLE)(-1)) {
        printf("Error: can't open driver\n");
        exit(1);
    }

    if(!DeviceIoControl(h,IOCTL_DbgPrnHk_GetVersion,
                        NULL,0,
                        &VerBuffer, sizeof(VerBuffer),
                        &returned,NULL)) {
        printf("Error: incompatible driver version\n");
        exit(1);
    }
    if(VerBuffer.Major != PROD_VER_MJ ||
       VerBuffer.Minor != PROD_VER_MN ||
       VerBuffer.Sub   != PROD_VER_NSUB) {
        printf("Error: incompatible driver version\n");
        exit(1);
    }

    if(!stdout_mode) {
        // Init log-file name
        wcsncpy(LogFileN, LogFile, MAX_PATH-10);
        LogFileN[MAX_PATH-10] = 0;
        len = wcslen(LogFileN);

        NumPos = &LogFileN[len];
        LogFileSuf[0] = 0;
        for(i=len; i>0; i--) {
            a = LogFileN[i-1];
            if(a == '\\') {
                break;
            }
            if(a == '.') {
                NumPos = &LogFileN[i-1];
                wcscpy(LogFileSuf, NumPos);
                break;
            }
        }
    }

    // alloc buffers
    MsgBuffer = (PDbgPrnHk_GetMessages_USER_OUT)
        GlobalAlloc(GMEM_DISCARDABLE, BufferSize);
    if(!MsgBuffer) {
        printf("Error: can't allocate message buffer\n");
        exit(4);
    }
    MsgBuffer2 = (PDbgPrnHk_GetMessages_USER_OUT)
        GlobalAlloc(GMEM_DISCARDABLE, BufferSize);
    if(!MsgBuffer2) {
        printf("Error: can't allocate message buffer (2)\n");
        exit(4);
    }
    MsgBuffer2a = (PDbgPrnHk_GetMessages_USER_OUT)
        GlobalAlloc(GMEM_DISCARDABLE, BufferSize);
    if(!MsgBuffer2a) {
        printf("Error: can't allocate message buffer (2)\n");
        exit(4);
    }

    i = Start_i;

    printf(help_text2);

    while(!esc_key) {

        if(!stdout_mode) {
            // find unused name for log-file
            if(i>0) {
                swprintf(NumPos, L"%d%s", i, LogFileSuf);
            }
            fh = CreateFileW(LogFileN, GENERIC_READ | GENERIC_WRITE,
                           FILE_SHARE_READ | FILE_SHARE_WRITE,
                           NULL,
                           OPEN_EXISTING,
                           FILE_ATTRIBUTE_NORMAL/* | FILE_FLAG_NO_BUFFERING*/,  NULL);
            if(fh && fh != (HANDLE)(-1)) {
                CloseHandle(fh);
                fh = NULL;
                i++;
                if(!i) {
                    printf("Error: find unused filename\n");
                    exit(2);
                }
                continue;
            }
            if((i >= MaxLogs) && (i-MaxLogs >= Start_i)) {
                if(i-MaxLogs > 0) {
                    swprintf(NumPos, L"%d%s", i-MaxLogs, LogFileSuf);
                } else {
                    swprintf(NumPos, L"%s", LogFileSuf);
                }
                DeleteFileW(LogFileN);
                if(i > 0) {
                    swprintf(NumPos, L"%d%s", i, LogFileSuf);
                } else {
                    swprintf(NumPos, L"%s", LogFileSuf);
                }
            }
            // create log-file
            fh = CreateFileW(LogFileN, GENERIC_READ | GENERIC_WRITE,
                           FILE_SHARE_READ | FILE_SHARE_WRITE,
                           NULL,
                           CREATE_NEW,
                           FILE_ATTRIBUTE_NORMAL/* | FILE_FLAG_NO_BUFFERING*/,  NULL);
        } else {
            fh = StdOut;
        }
        if(!fh || fh == (HANDLE)(-1)) {
            printf("Error: can't create log-file\n");
            fh = NULL;
            exit(3);
        }
        // write log header
        new_line = TRUE;
        if(!stdout_mode) {
            printf("Writing log to %ws\n", LogFileN);
        }
        CurLogSize = 0;
        len = sprintf(DbgPrintBuf, "Log generated by DbgPrintLog v%d.%d%s\n", PROD_VER_MJ, PROD_VER_MN, PROD_VER_SUB);
        TeeOutput(fh, DbgPrintBuf, len);

        GetSystemTime(&SysTime);
        len = sprintf(DbgPrintBuf, "Timestamp: %d/%2.2d/%2.2d %2.2d:%2.2d:%2.2d\n",
            SysTime.wYear,
            SysTime.wMonth,
            SysTime.wDay,
            SysTime.wHour,
            SysTime.wMinute,
            SysTime.wSecond);
        TeeOutput(fh, DbgPrintBuf, len);

        OsVer.dwOSVersionInfoSize = sizeof(OsVer);
        GetVersionEx(&OsVer);
        switch(OsVer.dwPlatformId) {
        case VER_PLATFORM_WIN32s:
            PlatformStr = "Win";
            break;
        case VER_PLATFORM_WIN32_WINDOWS:
            PlatformStr = "Win 9x";
            break;
        case VER_PLATFORM_WIN32_NT:
            PlatformStr = "NT";
            break;
        default:
            PlatformStr = "Unknown";
        }
        len = sprintf(DbgPrintBuf, "OS Version: %d:%s %d.%d (%d)\n  %s\n",
            OsVer.dwPlatformId,
            PlatformStr,
            OsVer.dwMajorVersion,
            OsVer.dwMinorVersion,
            OsVer.dwBuildNumber,
            OsVer.szCSDVersion);
        TeeOutput(fh, DbgPrintBuf, len);

        next_log = FALSE;
        while(!esc_key && !next_log) {
            // Get kernel messages
            if(!DeviceIoControl(h,IOCTL_DbgPrnHk_GetMessages,
                                MsgBuffer, min(BufferSize, 0x10000),
                                MsgBuffer, min(BufferSize, 0x10000),
                                &returned,NULL)) {
//                esc_key = TRUE;
//                break;
                returned = GetLastError();
                returned = 0;
            }
            // Get user-mode messages
            WaitForSingleObject(hODS_lock, INFINITE);

            // switch read/write buffers
            tmp_MsgBuffer2 = MsgBuffer2;

            MsgCount2   = MsgCount2a;
            MsgBuffer2  = MsgBuffer2a;

            MsgCount2a  = 0;
            MsgBuffer2a = tmp_MsgBuffer2;

            ReleaseMutex(hODS_lock);

            if(skip_kernel) {
                returned = 0;
            }
            if(skip_umode) {
                MsgCount2 = 0;
            }
            // drop messages if paused
            if(pause_key) {
                returned = 0;
                MsgCount2 = 0;
            }
            if(!returned && !MsgCount2) {
                SleepCount++;
                if(SleepCount > MaxSleepCount) {
                    SleepCount = 0;
                    new_line = TRUE;
                    FlushFileBuffered(fh);
                }
                Sleep(1000);
            }
            SleepCount = 0;
            j = 0;
            j2 = 0;
            // join user- and kernel-mode logs
            while(returned >= sizeof(DbgPrnHk_GetMessages_USER_OUT) || MsgCount2) {
                // assure time ordering
                if(returned >= sizeof(DbgPrnHk_GetMessages_USER_OUT) && MsgCount2) {
                    if(MsgBuffer[j] .TimeStamp.QuadPart >
                       MsgBuffer2[j2].TimeStamp.QuadPart) {
                        CurMsg = &MsgBuffer2[j2];
                        if(!last2) {
                            new_line = TRUE;
                            last2 = TRUE;
                        }
                    } else {
                        CurMsg = &MsgBuffer[j];
                        if(last2) {
                            new_line = TRUE;
                            last2 = FALSE;
                        }
                    }
                } else
                if(returned >= sizeof(DbgPrnHk_GetMessages_USER_OUT)) {
                    CurMsg = &MsgBuffer[j];
                    if(last2) {
                        new_line = TRUE;
                        last2 = FALSE;
                    }
                } else {
                    CurMsg = &MsgBuffer2[j2];
                    if(!last2) {
                        new_line = TRUE;
                        last2 = TRUE;
                    }
                }
                // write message to file (buffered)
                if(CurMsg->Length) {
                    if(new_line) {
                        if(log_mode) {
                            TeeOutput(fh, last2 ? "U " : "K ", 2);
                            CurLogSize+=2;
                        }
                        if(log_irql) {
                            len = sprintf(DbgPrintBuf, "%2.2x ",
                                CurMsg->Irql);
                            TeeOutput(fh, DbgPrintBuf, 3);
                            CurLogSize+=3;
                        }
                        if(log_pid) {
                            len = sprintf(DbgPrintBuf, "%8.8x ",
                                CurMsg->ProcessId);
                            TeeOutput(fh, DbgPrintBuf, len);
                            CurLogSize+=len;
                        }
                        if(log_tid) {
                            if(CurMsg->ThreadId != (PVOID)(-1)) {
                                len = sprintf(DbgPrintBuf, "%8.8x ",
                                    CurMsg->ThreadId);
                            } else {
                                len = sprintf(DbgPrintBuf, "???????? ");
                            }
                            TeeOutput(fh, DbgPrintBuf, len);
                            CurLogSize+=len;
                        }
                        len = sprintf(DbgPrintBuf, "%I64d.%4.4d\t",
                            CurMsg->TimeStamp.QuadPart/10000,
                            (ULONG)(CurMsg->TimeStamp.QuadPart%10000));
                        TeeOutput(fh, DbgPrintBuf, len);
                        CurLogSize+=len;
                        new_line = FALSE;
                    }
                    len = CurMsg->Length;
                    if(CurMsg->Msg[len-1] == '\n') {
                        new_line = TRUE;
                    }
                    TeeOutput(fh, &CurMsg->Msg[0], len);
                    CurLogSize+=len;
                }
                // go to next message
                if(last2) {
                    j2++;
                    MsgCount2--;
                } else {
                    j++;
                    returned -= sizeof(DbgPrnHk_GetMessages_USER_OUT);
                }
            }
            if(j+j2 < 256) {
                Sleep(100);
            }
            // check user controls
            if(flush_file || sync_mode) {
                if(!stdout_mode && !sync_mode) {
                    printf("flush log\n");
                }
                FlushFileBuffered(fh);
                flush_file = FALSE;
            }
            if(next_log) {
                printf("next log\n");
            }
            if(CurLogSize > MaxLogSize && !stdout_mode) {
                next_log = TRUE;
                printf("max log size limit reached -> next log\n");
            }
        } // end of Get Message loop
        FlushFileBuffered(fh);
        CloseHandle(fh);
        fh = NULL;
    }

    // cleanup
    if(fh) {
        CloseHandle(fh);
    }
    CloseHandle(h);
} // end get_messages()

static const char help_text[] = 
    "Usage:\n"
    "\tDbgPrintLog.exe\t[<switches>] [<log-file name>]\n"
    "Switches:\n"
    "\t-s NUM   set max log size to NUM Mbytes (0 < NUM < 2048)\n"
    "\t-l NUM   keep NUM latest logs\n"
    "\t-n NUM   set start log number to NUM\n"
    "\t-ft NUM  flush messages buffer after NUM seconds of inactivity\n"
    "\t-sm      synchronous mode. all messages are flushed immediately\n"
    "\t-o\n"
    "\t   or\n"
    "\t--stdout write log to STDOUT istead of file\n"
    "\t-cf      copy log to both STDOUT and file\n"
    "\t-p       log ProcessId\n"
    "\t-t       log ThreadId (for kernel-mode only)\n"
    "\t-m       log mode (kernel/user)\n"
//    "\t+fp PID  log only messages with ProcessId==PID\n"
//    "\t-fp PID  do not log messages with ProcessId==PID\n"
    "\t-fm M    do not log messages from M mode (M can be K or U)\n"
    "\t-h       display this help message\n"
    ;

void
usage(void)
{
    HANDLE stdOuth = GetStdHandle(STD_OUTPUT_HANDLE);
    ULONG WrittenBytes;

    if(stdOuth && stdOuth != ((HANDLE)(-1))) {
        WriteFile(stdOuth, help_text,  strlen(help_text),  &WrittenBytes, NULL);
        WriteFile(stdOuth, help_text2, strlen(help_text2), &WrittenBytes, NULL);
    } else {
        MessageBox(NULL, help_text, "Usage", MB_OK);
    }
}

void main () {
    int argc;
    WCHAR** argv;
    WCHAR*  CmdLine;

    LONG i=0;
    ULONG k;
//    WCHAR a;
    ULONG thId;

    printf("DbgPrintLog v%d.%d%s (c) by Alter\nHome site http://www.alter.org.ua\n",
        PROD_VER_MJ, PROD_VER_MN, PROD_VER_SUB);
    fflush(stdout);

    CmdLine = GetCommandLineW();
    argv = CommandLineToArgvW(CmdLine, &argc);

/*    if(argc >= 2) {
        LogFile = argv[1];
    }*/

    for(i=1; i<argc; i++) {
        if(!argv[i] || argv[i][0] != '-') {
            LogFile = argv[i];
            continue;
        }
        if(!wcscmp(argv[i], L"-h") ||
           !wcscmp(argv[i], L"--help")) {
            usage();
            exit(0);
        } else
        if(!wcscmp(argv[i], L"-sm")) {
            sync_mode = TRUE;
        } else
        if(!wcscmp(argv[i], L"-s")) {
            i++;
            if(i >= argc) {
                usage();
                exit(-2);
            }
            swscanf(argv[i], L"%d", &MaxLogSize);
            if(!MaxLogSize || MaxLogSize > 2047) {
                usage();
                exit(-2);
            }
            MaxLogSize = MaxLogSize << 20;
        } else
        if(!wcscmp(argv[i], L"-l")) {
            i++;
            if(i >= argc) {
                usage();
                exit(-2);
            }
            swscanf(argv[i], L"%d", &MaxLogs);
        } else
        if(!wcscmp(argv[i], L"-n")) {
            i++;
            if(i >= argc) {
                usage();
                exit(-2);
            }
            swscanf(argv[i], L"%d", &Start_i);
        } else
        if(!wcscmp(argv[i], L"-ft")) {
            i++;
            if(i >= argc) {
                usage();
                exit(-2);
            }
            swscanf(argv[i], L"%d", &MaxSleepCount);
        } else
        if(!wcscmp(argv[i], L"-fm")) {
            i++;
            if(i >= argc) {
                usage();
                exit(-2);
            }
            if(!wcsicmp(argv[i], L"K")) {
                skip_kernel = TRUE;
            } else
            if(!wcsicmp(argv[i], L"U")) {
                skip_umode = TRUE;
            } else {
                usage();
                exit(-2);
            }
        } else
        if(!wcscmp(argv[i], L"-cf")) {
            stdout_copy = TRUE;
            stdout_mode = FALSE;
        } else
        if(!wcscmp(argv[i], L"--stdout") ||
           !wcscmp(argv[i], L"-o")) {
            stdout_mode = TRUE;
            stdout_copy = FALSE;
        } else
        if(!wcscmp(argv[i], L"-i")) {
            log_irql = TRUE;
        } else
        if(!wcscmp(argv[i], L"-p")) {
            log_pid = TRUE;
        } else
        if(!wcscmp(argv[i], L"-t")) {
            log_tid = TRUE;
        } else
        if(!wcscmp(argv[i], L"-m")) {
            log_mode = TRUE;
        }
    }

    StdIn  = GetStdHandle(STD_INPUT_HANDLE);
    StdOut = GetStdHandle(STD_OUTPUT_HANDLE);

    if(!Privilege(SE_LOAD_DRIVER_NAME, TRUE)) {
        printf("Warning: Insufficient privileges\n");
    }

    // try to install (or re-install) and start kernel-mode DbgPrint capturing driver
    k = NtServiceStop(NT_DbgPrnHk_SVC_NAME);

    k = NtServiceInstall(NT_DbgPrnHk_SVC_NAME, ".");

    k = NtServiceIsRunning(NT_DbgPrnHk_SVC_NAME);
    if(k == 1) {
        // Ok
    } else
    if(k == 0) {
        k = NtServiceStart(NT_DbgPrnHk_SVC_NAME);
    }
    if(k < 0) {
        printf("Error: Can't load or install DbgPrnHk driver\n");
        exit(-1);
    }

    // init lock for switching buffers
    hODS_lock = CreateMutex(NULL, FALSE, NULL);
    if(!hODS_lock || hODS_lock == (HANDLE)(-1)) {
        printf("can't init mutex, (%#x)\n",
            GetLastError());
        exit(-1);
    }

    // start user-mode OutputDebugString capturing thread
    CreateThread(NULL, 0, get_ODS_messages, NULL, 0, &thId);
    // start keyboard event handling thread
    CreateThread(NULL, 0, get_kbd, NULL, 0, &thId);
    // get messages from user- and kernel-mode and store them to log-file
    get_messages();

    exit(0);
} // end main()
